# Copyright (c) 2022-2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
from .lorentz import Lorentzian
from .xy_series import XYSeries
from .polynomial_fit import PolynomialFit1D, PolynomialFit2D


__all__ = ['Lorentzian',
           'PolynomialFit1D',
           'PolynomialFit2D',
           'XYSeries',
           ]



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
from .config import Config, InvalConfigException


__all__ = ['Config',
           'InvalConfigException',
           ]



import numpy as np


class XYSeries:
    def __init__(self, X, Y):
        assert len(X) == len(Y)
        self.X = np.array(X)
        self.Y = np.array(Y)

    def _integrate(self, x0, x1):
        assert x0 <= x1

        x0, x1 = coords = np.clip([x0, x1], self.X[0], self.X[-1])
        yL, yH = np.interp(coords, self.X, self.Y)
        iL, iH = np.searchsorted(self.X, coords)
        X      = np.concatenate(([x0], self.X[iL:iH], [x1]))
        Y      = np.concatenate(([yL], self.Y[iL:iH], [yH]))
        A      = np.trapz(Y, x=X)

        return A, x0, x1

    def integrate(self, x0, x1):
        '''
        Integrate the data over the range [x0, x1] using the trapezoidal rule.
        The endpoints need not coincide with the sample data; the integration
        will use linear interpolation to integrate just the part of the
        boundary trapezoid that lies within the range.
        '''
        if x0 <= x1:
            return self._integrate(x0, x1)[0]
        return -self._integrate(x1, x0)[0]  # pylint: disable=W1114

    def get_avg_value(self, x0, x1):
        '''
        Compute the average value of the data over the range [x0, x1].  If the
        range extends past the endpoints of the data set, the x0 and x1 values
        will be clipped to the data set endpoints before computing the average
        value.  If the average value is undefined (x0 == x1 after clipping)
        then None is returned.
        '''
        area, x0, x1 = self._integrate(x0, x1)
        if x0 != x1:
            return area / (x1 - x0)
        return None

    def interpolate(self, x):
        '''
        Given an X value, interpolate what the Y value would be.  The X value
        is clipped to the range of the series.
        '''
        return np.interp(x, self.X, self.Y)

    def append(self, x, y):
        self.X = np.append(self.X, x)
        self.Y = np.append(self.Y, y)


if __name__ == '__main__':
    s = XYSeries([1, 2, 3], [4, 5, 6])
    assert s.integrate(1, 1) == 0
    assert s.integrate(0, 1) == 0
    assert s.integrate(0, 0.5) == 0
    assert s.integrate(3, 4) == 0
    assert s.integrate(3.5, 4) == 0
    assert s.integrate(0, 4) == 10
    assert s.integrate(1, 2) == 4.5
    assert s.integrate(1, 3) == 10
    assert s.integrate(2, 3) == 5.5
    assert s.integrate(1.5, 2.5) == 5
    assert s.integrate(1.5, 1.5) == 0

    s = XYSeries([0, 2, 3, 4], [2, 3, 5, 4])
    assert s.integrate(-1, 3) == 9
    assert s.integrate(-1, -0.5) == 0
    assert s.integrate(3, 5) == 4.5
    assert s.integrate(4.5, 5) == 0
    assert s.integrate(-1, 5) == 13.5
    assert s.integrate(0, 1) == 2.25
    assert s.integrate(0, 2) == 5
    assert s.integrate(0, 3) == 9
    assert s.integrate(0, 4) == 13.5
    assert s.integrate(0.5, 1.5) == 2.5
    assert s.integrate(0.5, 2.5) == 5.6875
    assert s.integrate(2.5, 0.5) == -5.6875

    print('Tests succeeded.')



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse

import xtalx.z_sensor


def main(args):
    dev = xtalx.z_sensor.find_one(serial_number=args.sensor)
    tc  = xtalx.z_sensor.make(dev, yield_Y=True)
    tc.info('Stopping sensor...')
    tc.send_scope_cmd(32768, 0)
    tc.info('Stopped.')


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--sensor', '-s')
    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    _main()



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import time

import xtalx.z_sensor

from . import z_common


def main(rv):
    dev    = xtalx.z_sensor.find_one(serial_number=rv.sensor)
    tc     = xtalx.z_sensor.make(dev, verbose=rv.verbose,
                                 yield_Y=not rv.track_impedance)
    za, zl = z_common.parse_args(tc, rv)
    pt     = xtalx.z_sensor.PeakTracker(tc, za.amplitude, za.f0, za.f1, za.df,
                                        za.nfreqs, za.search_time_secs,
                                        za.sweep_time_secs,
                                        settle_ms=za.settle_ms,
                                        delegate=z_common.ZDelegate(zl))
    pt.start_threaded()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print()
    finally:
        pt.stop_threaded()


def _main():
    parser = argparse.ArgumentParser()
    z_common.add_arguments(parser)
    rv = parser.parse_args()
    main(rv)


if __name__ == '__main__':
    _main()



# Copyright (c) 2022-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import usb.core
import usb.backend.libusb1
import libusb_package


LIBUSB_BACKEND = None


def get_backend():
    global LIBUSB_BACKEND
    if not LIBUSB_BACKEND:
        LIBUSB_BACKEND = usb.backend.libusb1.get_backend(
                find_library=libusb_package.find_library)
    return LIBUSB_BACKEND


def find(serial_number=None, **kwargs):
    opt_args = {}
    if serial_number is not None:
        opt_args['serial_number'] = serial_number
    return list(usb.core.find(**kwargs, **opt_args, backend=get_backend()))


def find_one(**kwargs):
    usb_devs = find(**kwargs)
    if len(usb_devs) > 1:
        raise Exception('Multiple matching devices: %s' %
                        ', '.join(ud.serial_number for ud in usb_devs))
    if not usb_devs:
        raise Exception('No matching devices.')
    return usb_devs[0]


__all__ = ['find',
           'get_backend',
           ]



#!/usr/bin/env python3
# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import time

import xtalx.z_sensor


def date_str(posix_time):
    return '%s (%s)' % (posix_time, time.asctime(time.localtime(posix_time)))


def print_info_tcsc_u5(tc2):
    print('     Serial Num: %s' % tc2.serial_num)
    print('       Git SHA1: %s' % tc2.git_sha1)
    print('     FW Version: 0x%04X' % tc2.fw_version)
    print('           HCLK: %u' % tc2.ginfo.hclk)
    print('           DCLK: %u' % tc2.ginfo.dclk)
    print('           ACLK: %u' % tc2.ginfo.aclk)
    print('            HSE: %u MHz' % tc2.ginfo.f_hs_mhz)
    print('DV Nominal Freq: %u Hz' % tc2.ginfo.dv_nominal_hz)
    print('    Max cmd len: %u' % tc2.ginfo.cmd_buf_len)
    print('       # resets: %u' % tc2.ginfo.nresets)
    print('     Drive Type: %s' % tc2.ginfo.get_drive_type())
    print('   Reset Reason: %s' % tc2.ginfo.reset_reason)
    print('      Reset CSR: 0x%08X' % tc2.ginfo.reset_csr)
    print('      Reset SR1: 0x%08X' % tc2.ginfo.reset_sr1)
    print('Max sweep freqs: %u' % tc2.ginfo.max_sweep_entries)

    print('-----------------------')
    if not tc2.ginfo.electronics_cal_date:
        print('No electrical calibration present.')
    else:
        vdda = tc2.adc_to_v(tc2.ADC_MAX / 2)
        print('Electrical calibration:')
        print(' Calibration Date: %s' %
              date_str(tc2.ginfo.electronics_cal_date))
        print('         R Source: %f' % tc2.einfo.r_source)
        print('       R Feedback: %f' % tc2.einfo.r_feedback)
        print('         DAC-to-V: %f + %f*x' % (tc2.einfo.dac_to_v_coefs[0],
                                                tc2.einfo.dac_to_v_coefs[1]))
        print('         ADC-to-V: %f + %f*x' % (tc2.einfo.adc_to_v_coefs[0],
                                                tc2.einfo.adc_to_v_coefs[1]))
        print('     Nominal VDDA: %.2f' % vdda)

        a_dac = tc2.dac_to_a(tc2.DAC_MAX)
        print('               G0: %f' % (a_dac / vdda))

    print('-----------------------')
    if not tc2.ginfo.crystal_cal_date:
        print('No tuning fork calibration present.')
    else:
        print('Tuning fork calibration:')
        print(' Calibration Date: %s' % date_str(tc2.ginfo.crystal_cal_date))
    if tc2.ginfo.cal_dac_amplitude:
        v = tc2.dac_to_a(tc2.ginfo.cal_dac_amplitude)
        if v is not None:
            print('Cal DAC Amplitude: %u (%.2fV)' %
                  (tc2.ginfo.cal_dac_amplitude, v))
        else:
            print('Cal DAC Amplitude: %u' % tc2.ginfo.cal_dac_amplitude)
    else:
        print('Cal DAC Amplitude: Not present')


def main(rv):
    devs = xtalx.z_sensor.find(serial_number=rv.sensor)
    for dev in devs:
        tc2 = xtalx.z_sensor.make(dev)
        print('*************************************')
        if isinstance(tc2, xtalx.z_sensor.TCSC_U5):
            print_info_tcsc_u5(tc2)
            continue

        print('Unrecognized TinCan type.')


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--sensor')
    rv = parser.parse_args()

    try:
        main(rv)
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import threading
import random
import errno
import time
from enum import IntEnum

import usb
import usb.util

import btype

from .exception import XtalXException


FC_FLAGS_VALID              = (1 << 15)
FC_FLAG_NO_TEMP_PRESSURE    = (1 << 4)
FC_FLAG_PRESSURE_FAILED     = (1 << 3)
FC_FLAG_TEMP_FAILED         = (1 << 2)
FC_FLAG_PRESSURE_UPDATE     = (1 << 1)
FC_FLAG_TEMP_UPDATE         = (1 << 0)


class Status(IntEnum):
    OK              = 0
    BAD_LENGTH      = 1
    BAD_OPCODE      = 2
    NOT_FOUND       = 3

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            s = '%s' % Status(rsp.status)
        except ValueError:
            s = '%u' % rsp.status
        return s


class CommandException(Exception):
    def __init__(self, rsp):
        super().__init__(
            'Command exception: %s (%s)' % (Status.rsp_to_status_str(rsp), rsp))
        self.rsp = rsp


class Opcode(IntEnum):
    GET_INFO        = 1
    GET_PID_INFO    = 2
    SET_PID_INFO    = 3
    DISABLE_PID     = 4
    ENABLE_PID      = 5
    SET_DAC_VALUE   = 6
    BAD_OPCODE      = 0xCCCC


class CommandHeader(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    flags           = btype.uint32_t()
    _EXPECTED_SIZE  = 8


class SetPIDInfoPayload(btype.Struct):
    Kp              = btype.float64_t()
    Ki              = btype.float64_t()
    Kd              = btype.float64_t()
    setpoint_c      = btype.float64_t()
    _EXPECTED_SIZE  = 32


class SetDACValuePayload(btype.Struct):
    dac_value       = btype.float64_t()
    _EXPECTED_SIZE  = 8


class Response(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    params          = btype.Array(btype.uint32_t(), 12)
    _EXPECTED_SIZE  = 56


class GetInfoResponse(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    flags           = btype.uint32_t()
    f_hs_mhz        = btype.uint32_t()
    hclk            = btype.float64_t()
    _EXPECTED_SIZE  = 24


class GetPIDInfoResponse(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    Kp              = btype.float64_t()
    Ki              = btype.float64_t()
    Kd              = btype.float64_t()
    setpoint_c      = btype.float64_t()
    dac_t_val       = btype.float64_t()
    _EXPECTED_SIZE  = 48


class FrequencyPacket24(btype.Struct):
    '''
    Firmware revisions 1.0.6 and earlier return a 24-byte packet if the sensor
    doesn't have enough data to perform a temperature-compensated pressure
    measurement yet or if the sensor doesn't have a calibration applied in
    flash.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    _EXPECTED_SIZE      = 24


class FrequencyPacket40(btype.Struct):
    '''
    Firmware revisions 1.0.6 and earlier return a 40-byte packet if the sensor
    has enough data to perform a temperature-compensated pressure measurement.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    pressure_psi        = btype.float64_t()
    temp_c              = btype.float64_t()
    _EXPECTED_SIZE      = 40


class FrequencyPacket56(btype.Struct):
    '''
    Firmware revisions 1.0.7 and higher always return a 56-byte packet that
    contains flags indicating the validity of things like the temperature-
    compensated pressure measurement.  These firmware versions also return the
    MCU temperature as a control.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    pressure_psi        = btype.float64_t()
    temp_c              = btype.float64_t()
    mcu_temp_c          = btype.float64_t()
    rsrv2               = btype.Array(btype.uint8_t(), 8)
    _EXPECTED_SIZE      = 56


class FrequencyPacket56_110(btype.Struct):
    '''
    Firmware revision 1.1.0 redoes the entire packet contents, while
    maintaining backwards-compatibility for the essential fields.
    '''
    cP                     = btype.float32_t()
    cI                     = btype.float32_t()
    cD                     = btype.float32_t()
    dac_2p20               = btype.uint32_t()
    pressure_hz_1e4        = btype.uint32_t()
    flags                  = btype.uint16_t()
    seq_num                = btype.uint8_t()
    rsrv1                  = btype.uint8_t()
    pressure_psi           = btype.float64_t()
    temp_c                 = btype.float64_t()
    temp_hz_1e4            = btype.uint32_t()
    lores_pressure_hz_1e4  = btype.uint32_t()
    lores_temp_hz_1e4      = btype.uint32_t()
    lores_pressure_psi     = btype.float32_t()
    lores_temp_c           = btype.float32_t()
    _EXPECTED_SIZE         = 60


class Measurement:
    '''
    Object encapsulating the results of an XTI sensor measurement.  The
    following fields are defined:

        sensor - Reference to the XTI that generated the Measurement.
        ref_freq - Frequency of the sensor's reference crystal.
        pressure_edges - Number of pressure crystal ticks used to generate the
            Measurement.
        pressure_ref_clocks - Number of reference clock ticks that elapsed
            while counting pressure_edges pressure crystal ticks.
        pressure_freq - Measured pressure crystal frequency.
        temp_edges - Number of temperature crystal ticks used to generate the
            Measurement.
        temp_ref_clocks - Number of temperature crystal ticks that elapsed
            while counting temp_edges temperature crystal ticks.
        temp_freq - Measured temperature crystal frequency.
        mcu_temp_c - Microcontroller's internal junction temperature.
        pressure_psi - Temperature-compensated pressure measured in PSI.
        temp_c - Temperature measured in degrees Celsius.
        flags - A set of validity and error flags.

    If the sensor is uncalibrated or has not sampled enough data to generate
    a temperature-compensated pressure measurement then some or all of
    temp_freq, pressure_freq, pressure_psi and temp_c may be None.

    The flags field is a bitmask which may include any of the following bits;
    it may be None if the firmware version predates the introduction of status
    flags:

        FC_FLAGS_VALID - The flags field contains valid information (always set
            or flags will be None).
        FC_FLAG_NO_TEMP_PRESSURE - Will be set if pressure_psi and temp_c could
            not be generated; the sensor may be uncalibrated or may not have
            generated both temperature and pressure crystal readings yet.
        FC_FLAG_PRESSURE_FAILED - Will be set if 0.5 seconds elapse without a
            pressure crystal measurement completing; this indicates that a
            sensor failure has caused the pressure crystal to stop ticking.
        FC_FLAG_TEMP_FAILED - Will be set if 0.5 seconds elapse without a
            temperature crystal measurement completing; this indicates that a
            sensor failure has caused the temperature crystal to stop ticking.
        FC_FLAG_PRESSURE_UPDATE - Indicates that the current Measurement
            incorporates a new reading from the pressure crystal; it may still
            be incorporating the previous reading from the temperature crystal.
        FC_FLAG_TEMP_UPDATE - Indicates that the current Measurement
            incorporates a new reading from the temperature crystal; it may
            still be incorporating the previous reading from the pressure
            crystal.

    Note that since the temperature and pressure crystals tick asynchronously
    with respect to one another, a measurement on one crystal is likely to
    complete while a measurement on the other crystal is still pending and so
    typically only one of FC_FLAG_PRESSURE_UPDATE or FC_FLAG_TEMP_UPDATE will
    be set.
    '''
    def __init__(self, sensor, mcu_temp_c, pressure_psi, temp_c, pressure_freq,
                 temp_freq, lores_pressure_psi, lores_temp_c,
                 lores_pressure_freq, lores_temp_freq, cP, cI, cD, dac, flags):
        self.sensor              = sensor
        self.mcu_temp_c          = mcu_temp_c
        self.pressure_psi        = pressure_psi
        self.temp_c              = temp_c
        self.pressure_freq       = pressure_freq
        self.temp_freq           = temp_freq
        self.lores_pressure_psi  = lores_pressure_psi
        self.lores_temp_c        = lores_temp_c
        self.lores_pressure_freq = lores_pressure_freq
        self.lores_temp_freq     = lores_temp_freq
        self.cP                  = cP
        self.cI                  = cI
        self.cD                  = cD
        self.dac                 = dac
        self.flags               = flags

    @staticmethod
    def _from_packet(sensor, packet):
        mt, p, t, Fp, Ft = None, None, None, None, None
        lp, lt, Flp, Flt = None, None, None, None
        cP, cI, cD, dac  = None, None, None, None

        if sensor.usb_dev.bcdDevice < 0x0107:
            if len(packet) == 24:
                fp = FrequencyPacket24.unpack(packet)
            else:
                fp = FrequencyPacket40.unpack(packet)
                p  = fp.pressure_psi
                t  = fp.temp_c
                Fp = fp.ref_freq*fp.pressure_edges/fp.pressure_ref_clocks
                Ft = fp.ref_freq*fp.temp_edges/fp.temp_ref_clocks
        elif sensor.usb_dev.bcdDevice < 0x0110:
            fp = FrequencyPacket56.unpack(packet)
            mt = fp.mcu_temp_c
            assert fp.flags and (fp.flags & FC_FLAGS_VALID)
            if not (fp.flags & FC_FLAG_PRESSURE_FAILED):
                Fp = fp.ref_freq*fp.pressure_edges/fp.pressure_ref_clocks
            if not (fp.flags & FC_FLAG_TEMP_FAILED):
                Ft = fp.ref_freq*fp.temp_edges/fp.temp_ref_clocks
            if (fp.flags & FC_FLAG_NO_TEMP_PRESSURE) == 0:
                p = fp.pressure_psi
                t = fp.temp_c
        else:
            fp = FrequencyPacket56_110.unpack(packet)
            assert fp.flags and (fp.flags & FC_FLAGS_VALID)
            if not (fp.flags & FC_FLAG_PRESSURE_FAILED):
                Fp  = fp.pressure_hz_1e4 / 1e4
                Flp = fp.lores_pressure_hz_1e4 / 1e4
            if not (fp.flags & FC_FLAG_TEMP_FAILED):
                Ft  = fp.temp_hz_1e4 / 1e4
                Flt = fp.lores_temp_hz_1e4 / 1e4
            if (fp.flags & FC_FLAG_NO_TEMP_PRESSURE) == 0:
                p   = fp.pressure_psi
                t   = fp.temp_c
                lp  = fp.lores_pressure_psi
                lt  = fp.lores_temp_c
                if sensor.is_pid_supported():
                    cP  = fp.cP
                    cI  = fp.cI
                    cD  = fp.cD
                    dac = fp.dac_2p20 / 2**20
        flags = fp.flags if fp.flags & FC_FLAGS_VALID else None

        return Measurement(sensor, mt, p, t, Fp, Ft, lp, lt, Flp, Flt, cP, cI,
                           cD, dac, flags)

    def tostring(self, verbose=False):
        s = '%s: ' % self.sensor
        if verbose:
            s += ('F 0x%04X pf %s tf %s p %s t %s lpf %s ltf %s lp %s lt %s '
                  'mt %s' %
                  (self.flags, self.pressure_freq, self.temp_freq,
                   self.pressure_psi, self.temp_c, self.lores_pressure_freq,
                   self.lores_temp_freq, self.lores_pressure_psi,
                   self.lores_temp_c, self.mcu_temp_c))
        else:
            if self.pressure_psi is None:
                p = 'n/a'
            else:
                p = '%f' % self.pressure_psi

            if self.temp_c is None:
                t = 'n/a'
            else:
                t = '%f' % self.temp_c
            s += '%s PSI, %s C' % (p, t)

        return s

    def to_influx_point(self, time_ns=None, measurement='xtalx_data',
                        fields=None):
        time_ns = time_ns or time.time_ns()
        fields  = fields or {}
        p = {
            'measurement' : measurement,
            'time'        : time_ns,
            'tags'        : {'sensor' : self.sensor.serial_num},
            'fields'      : fields,
        }
        if self.mcu_temp_c is not None:
            fields['mcu_temp_c'] = float(self.mcu_temp_c)
        if self.pressure_psi is not None:
            fields['pressure_psi'] = float(self.pressure_psi)
        if self.temp_c is not None:
            fields['temp_c'] = float(self.temp_c)
        if self.pressure_freq is not None:
            fields['pressure_freq_hz'] = float(self.pressure_freq)
        if self.temp_freq is not None:
            fields['temp_freq_hz'] = float(self.temp_freq)
        if self.lores_pressure_psi is not None:
            fields['lores_pressure_psi'] = float(self.lores_pressure_psi)
        if self.lores_temp_c is not None:
            fields['lores_temp_c'] = float(self.lores_temp_c)
        if self.lores_pressure_freq is not None:
            fields['lores_pressure_freq_hz'] = float(self.lores_pressure_freq)
        if self.lores_temp_freq is not None:
            fields['lores_temp_freq_hz'] = float(self.lores_temp_freq)
        if self.cP is not None:
            fields['cP']  = float(self.cP)
            fields['cI']  = float(self.cI)
            fields['cD']  = float(self.cD)
            fields['dac'] = float(self.dac)
        return p


class XTI:
    TELEMETRY_EP    = 0x81
    CMD_EP          = 0x02
    RSP_EP          = 0x83

    '''
    Given a USB device handle acquired via find() or find_one(), creates an
    XTI object that can be used to communicate with a sensor.
    '''
    def __init__(self, usb_dev):
        self.usb_dev     = usb_dev
        self.lock        = threading.RLock()
        self._halt_yield = True
        self.thread      = None

        try:
            self.serial_num = usb_dev.serial_number
            self.git_sha1   = usb.util.get_string(usb_dev, 6)
            self.fw_version = usb_dev.bcdDevice
        except ValueError as e:
            if str(e) == 'The device has no langid':
                raise XtalXException(
                    'Device has no langid, ensure running as root!') from e

        if self.usb_dev.bcdDevice >= 0x0103:
            try:
                self.report_id = int(usb.util.get_string(usb_dev, 15))
            except ValueError:
                self.report_id = None
        else:
            self.report_id = None

        if self.usb_dev.bcdDevice >= 0x0110:
            self.tag = random.randint(1, 0xFFFF)
            self._set_measurement_config()
            self.xinfo = self._get_info()
            if self.xinfo.flags & (1 << 0):
                self.pinfo = self._get_pid_info()
            else:
                self.pinfo = None

        self.usb_path = '%s:%s' % (
            usb_dev.bus, '.'.join('%u' % n for n in usb_dev.port_numbers))

    def __str__(self):
        return 'XTI(%s)' % self.serial_num

    def _set_configuration(self, bConfigurationValue):
        with self.lock:
            cfg = None
            try:
                cfg = self.usb_dev.get_active_configuration()
            except usb.core.USBError as e:
                if e.strerror != 'Configuration not set':
                    raise

            if cfg is None or cfg.bConfigurationValue != bConfigurationValue:
                usb.util.dispose_resources(self.usb_dev)
                self.usb_dev.set_configuration(bConfigurationValue)

    def _set_measurement_config(self):
        self._set_configuration(2)

    def _alloc_tag(self):
        tag      = self.tag
        self.tag = 1 if self.tag == 0xFFFF else self.tag + 1
        return tag

    def _send_command(self, opcode, flags, params, bulk_data, timeout):
        tag  = self._alloc_tag()
        hdr  = CommandHeader(opcode=opcode, tag=tag, flags=flags)
        data = hdr.pack() + params + bytes(48 - len(params)) + bulk_data
        size = self.usb_dev.write(self.CMD_EP, data, timeout=timeout)
        assert size == len(data)
        return tag

    def _recv_response(self, tag, timeout, cls=Response):
        data = self.usb_dev.read(self.RSP_EP, Response._STRUCT.size,
                                 timeout=timeout)
        assert len(data) == Response._STRUCT.size
        rsp = cls.unpack_from(data)
        assert rsp.tag == tag

        if rsp.status != Status.OK:
            rsp.opcode = Opcode(rsp.opcode)
            raise CommandException(rsp)

        return rsp

    def _exec_command(self, opcode, flags=0, params=b'', bulk_data=b'',
                      timeout=1000, cls=Response):
        tag = self._send_command(opcode, flags, params, bulk_data, timeout)
        return self._recv_response(tag, timeout, cls=cls)

    def _get_info(self):
        return self._exec_command(Opcode.GET_INFO, cls=GetInfoResponse)

    def _get_pid_info(self):
        return self._exec_command(Opcode.GET_PID_INFO, cls=GetPIDInfoResponse)

    def _set_pid_info(self, Kp=None, Ki=None, Kd=None, setpoint_c=None):
        flags = 0
        if Kp is not None:
            flags |= (1 << 0)
        if Ki is not None:
            flags |= (1 << 1)
        if Kd is not None:
            flags |= (1 << 2)
        if setpoint_c is not None:
            flags |= (1 << 3)
        payload = SetPIDInfoPayload(Kp=(Kp or 0), Ki=(Ki or 0), Kd=(Kd or 0),
                                    setpoint_c=(setpoint_c or 0))
        return self._exec_command(Opcode.SET_PID_INFO, flags, payload.pack())

    def _disable_pid(self):
        return self._exec_command(Opcode.DISABLE_PID)

    def _enable_pid(self):
        return self._exec_command(Opcode.ENABLE_PID)

    def _set_dac_value(self, dac_value):
        return self._exec_command(
                Opcode.SET_DAC_VALUE,
                params=SetDACValuePayload(dac_value=dac_value).pack())

    def is_pid_supported(self):
        return self.pinfo is not None

    def read_measurement(self, timeout=2000):
        '''
        Synchronously read a single measurement from the sensor, blocking if no
        measurement is currently available.
        '''
        with self.lock:
            p = self.usb_dev.read(self.TELEMETRY_EP, 64, timeout=timeout)
        return Measurement._from_packet(self, p)

    def _yield_measurements(self, do_reset, timeout):
        with self.lock:
            if do_reset:
                self.usb_dev.reset()
            self._set_measurement_config()

            while not self._halt_yield:
                try:
                    yield self.read_measurement(timeout=timeout)
                except usb.core.USBError as e:
                    if e.errno != errno.ETIMEDOUT:
                        raise
                    continue

    def yield_measurements(self, do_reset=True, timeout=2000):
        '''
        Yields Measurement objects synchronously in the current thread,
        blocking while waiting for new measurements to be acquired.
        '''
        with self.lock:
            self._halt_yield = False
            yield from self._yield_measurements(do_reset, timeout=timeout)

    def halt_yield(self):
        '''
        Halts an ongoing yield_measurements() call, causing it to eventually
        terminate the generator loop.
        '''
        self._halt_yield = True

    def _read_measurements_async(self, handler, do_reset, timeout):
        with self.lock:
            for m in self._yield_measurements(do_reset, timeout=timeout):
                handler(m)

    def read_measurements(self, handler, do_reset=True, timeout=2000):
        '''
        Reads measurements asynchronously in a separate thread, calling the
        handler as measurements become available.  The handler should take a
        single Measurement object as an argument.
        '''
        with self.lock:
            assert self.thread is None
            self._halt_yield = False
            self.thread = threading.Thread(target=self._read_measurements_async,
                                           args=(handler, do_reset, timeout),
                                           daemon=False)
            self.thread.start()

    def join_read(self):
        '''
        Blocks the current thread until the asynchronous read thread completes.
        Typically this blocks indefinitely until some error occurs, however the
        read thread will also exit if someone sets the _halt_yield field to
        True (see XTI.halt_read()).
        '''
        self.thread.join()

    def halt_read(self):
        '''
        Halts any asynchronous measurement thread and waits for it to finish
        cleanly.
        '''
        self._halt_yield = True
        self.join_read()



#!/usr/bin/env python3
# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import argparse
import usb.util

import xtalx.p_sensor


def main(_args):
    for s in xtalx.p_sensor.find_xti():
        print('******************')
        print('Sensor SN: %s' % s.serial_number)
        print(' git SHA1: %s' % usb.util.get_string(s, 6))
        print('  Version: 0x%04X' % s.bcdDevice)


def _main():
    parser = argparse.ArgumentParser()
    main(parser.parse_args())


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import argparse
import time
import math

import xtalx.p_sensor


def print_measurement(x, temp_c, pressure_psi):
    if pressure_psi is None:
        p = 'n/a'
    else:
        p = '%f' % pressure_psi

    if temp_c is None:
        t = 'n/a'
    else:
        t = '%f' % temp_c

    print('%s: %s C, %s PSI' % (x, t, p))


def xtalx_cb(x, t, dt, temp_c, pressure_psi, csv_file):
    print_measurement(x, temp_c, pressure_psi)
    if csv_file:
        if temp_c is None:
            temp_c = math.nan
        if pressure_psi is None:
            pressure_psi = math.nan
        csv_file.write('%.6f,%.6f,%.2f,%.5f\n' % (t, dt, temp_c, pressure_psi))
        csv_file.flush()


def sample_gated(x, csv_file, sample_period):
    measurements   = []
    t0 = t0_period = time.time()
    for m in x.yield_measurements():
        t = time.time()
        if t - t0_period < sample_period:
            measurements.append(m)
            continue

        if measurements:
            N      = len(measurements)
            temp_c = sum(m.temp_c for m in measurements) / N
            psi    = sum(m.pressure_psi for m in measurements) / N
            xtalx_cb(x, t, t - t0, temp_c, psi, csv_file)

        t0_period   += sample_period
        measurements = [m]


def sample_continuous(x, csv_file):
    t0 = time.time()
    for m in x.yield_measurements():
        t = time.time()
        xtalx_cb(x, t, t - t0, m.temp_c, m.pressure_psi, csv_file)


def main(args):
    if args.xhti_intf:
        x = xtalx.p_sensor.XHTI(args.xhti_intf)
    else:
        d = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
        x = xtalx.p_sensor.make(d)

    if args.csv_file:
        csv_file = open(  # pylint: disable=R1732
            args.csv_file, 'a+', encoding='utf8')

        pos = csv_file.tell()
        if pos != 0:
            csv_file.seek(0)
            if csv_file.read(28) != 'time,dt,temp_c,pressure_psi\n':
                print('%s does not appear to be a pressure sensor log file.' %
                      args.csv_file)
                return
            csv_file.seek(pos)
        else:
            csv_file.write('time,dt,temp_c,pressure_psi\n')
            csv_file.flush()
    else:
        csv_file = None

    if args.sample_period:
        sample_gated(x, csv_file, args.sample_period)
    else:
        sample_continuous(x, csv_file)


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    parser.add_argument('--xhti-intf')
    parser.add_argument('--csv-file')
    parser.add_argument('--sample-period', type=float)
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import serial


def from_intf(intf, spy=False, **kwargs):
    url = ('spy://%s' % intf) if spy else intf
    return serial.serial_for_url(url, **kwargs)



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import argparse
import cmath
import time
from enum import Enum

import glotlib
import glfw

import xtalx.z_sensor
from xtalx.z_sensor.peak_tracker import Delegate


# Search parameters.
PT_F0              = 20000
PT_F1              = 35000
PT_DF              = 50
PT_NFREQS          = 20
PT_SEARCH_TIME_SEC = 5
ENABLE_CHIRP       = False

# Sweep parameters.
PT_SWEEP_TIME_SEC  = 5
PT_SETTLE_TIME_MS  = 5000


class ViewMode(Enum):
    F_VS_T      = 1
    F_T_VS_TIME = 2


class WATWindow(glotlib.Window, Delegate):
    def __init__(self, tc, csv):
        super().__init__(1000, 700, msaa=4, name=tc.serial_num)

        self.tc             = tc
        self.csv            = csv
        self.data_lock      = threading.Lock()
        self.fits           = []
        self.points         = None
        self.auto_snap      = True
        self.end_sweep_time = None
        self.sweep_prefix   = ''
        self.view_mode      = ViewMode.F_VS_T

        self.f_vs_T_plot = self.add_plot((3, 4, (1, 11)),
                                         limits=(0, 27000, 130, 29500))
        self.f_vs_T_lines = self.f_vs_T_plot.add_lines(X=[], Y=[])
        self.f_vs_T_point = self.f_vs_T_plot.add_points(X=[], Y=[], width=3)
        self.f_vs_T_plot.set_x_label('Temperature (C)')
        self.f_vs_T_plot.set_y_label('Freq (Hz)')

        self.t_vs_time_plot = self.add_plot((2, 4, (1, 3)),
                                            limits=(-1, 0, 60, 130),
                                            visible=False)
        self.t_vs_time_lines = self.t_vs_time_plot.add_lines(X=[], Y=[])
        self.t_vs_time_plot.set_y_label('Temperature (C)')

        self.f_vs_time_plot = self.add_plot((2, 4, (5, 7)),
                                            limits=(-1, 27000, 60, 29500),
                                            visible=False,
                                            sharex=self.t_vs_time_plot)
        self.f_vs_time_lines = self.f_vs_time_plot.add_lines(X=[], Y=[])
        self.f_vs_time_plot.set_x_label('Time (min)')
        self.f_vs_time_plot.set_y_label('Freq (Hz)')

        self.rzx_plot = self.add_plot((3, 4, (4, 4)), max_h_ticks=3)
        self.rzx_lines = self.rzx_plot.add_lines(X=[], Y=[])
        self.rzx_plot.set_x_label('Freq (Hz)', side='top')
        self.rzx_plot.set_y_label('Real(Z)', side='right')

        self.phi_plot = self.add_plot((3, 4, (8, 8)), max_h_ticks=3,
                                      sharex=self.rzx_plot)
        self.phi_lines = self.phi_plot.add_lines(X=[], Y=[])
        self.phi_plot.set_y_label('Arg(Z)', side='right')

        self.nyq_plot = self.add_plot((3, 4, (12, 12)), max_h_ticks=3,
                                      aspect=glotlib.ASPECT_SQUARE)
        self.nyq_lines = self.nyq_plot.add_lines(X=[], Y=[])
        self.nyq_plot.set_x_label('Real(Z)')
        self.nyq_plot.set_y_label('-Imag(Z)', side='right')

        self.status_text = self.add_label((0.01, 0.01), '')
        self.pos_label = self.add_label((0.99, 0.01), '', anchor='SE')
        self.temp_label = self.add_label(
                (self.f_vs_T_plot.bounds[0] + 0.06,
                 self.f_vs_T_plot.bounds[3] + 0.01),
                '', anchor='SW')
        self.freq_label = self.add_label(
                (self.f_vs_T_plot.bounds[2] - 0.01,
                 self.f_vs_T_plot.bounds[3] + 0.01),
                '', anchor='SE')

        glotlib.periodic(1, self.update_periodic)

    def update_periodic(self, _t):
        self.mark_dirty()

    def handle_mouse_moved(self, _x, _y):
        self.mark_dirty()

    def handle_key_press(self, key):
        if key == glotlib.KEY_ESCAPE:
            self.swap_view()
        elif key == glfw.KEY_Z:
            self.default_zoom()

    def swap_view(self):
        if self.view_mode == ViewMode.F_VS_T:
            self.f_vs_T_plot.hide()
            self.f_vs_time_plot.show()
            self.t_vs_time_plot.show()
            self.view_mode = ViewMode.F_T_VS_TIME
        elif self.view_mode == ViewMode.F_T_VS_TIME:
            self.f_vs_T_plot.show()
            self.f_vs_time_plot.hide()
            self.t_vs_time_plot.hide()
            self.view_mode = ViewMode.F_VS_T

    def default_zoom(self):
        self.f_vs_T_plot._set_x_lim(0, 130)
        self.f_vs_T_plot._set_y_lim(27000, 29500)
        self.t_vs_time_plot._set_y_lim(0, 130)
        self.f_vs_time_plot._set_y_lim(27000, 29500)

    def set_status_text(self, t):
        return self.status_text.set_text('%s' % t)

    def update_geometry(self, _t):
        with self.data_lock:
            fits        = self.fits
            points      = self.points

            if fits:
                self.fits = []
            self.points = None

        updated = False

        _, _, _, data_x, data_y = self.get_mouse_pos()
        if data_x is not None:
            updated |= self.pos_label.set_text('%.4f  %.4f' %
                                               (data_x, data_y))

        if self.end_sweep_time is not None:
            dt = max(self.end_sweep_time - time.time(), 0)
            if dt == 0:
                updated |= self.set_status_text('%s: Fitting...' %
                                                self.sweep_prefix)
            else:
                updated |= self.set_status_text('%s: %.0f seconds' %
                                                (self.sweep_prefix, dt))
        else:
            updated |= self.set_status_text('')

        if fits:
            T = [fit._gl_dtime_ns / 60e9 for fit in fits]
            X = [fit.temp_c for fit in fits]
            Y = [fit.peak_hz for fit in fits]
            self.f_vs_T_lines.append_x_y_data(X, Y)
            self.f_vs_T_point.set_x_y_data(X=[X[-1]], Y=[Y[-1]])
            self.t_vs_time_lines.append_x_y_data(T, X)
            self.f_vs_time_lines.append_x_y_data(T, Y)
            self.temp_label.set_text('Temp: %.3fC' % X[-1])
            self.freq_label.set_text('Freq: %.3f Hz' % Y[-1])
            updated = True

        if points:
            X = [p.f for p in points]
            Y = [cmath.phase(p.z) for p in points]
            self.phi_lines.set_x_y_data(X, Y)

            Y = [p.z.real for p in points]
            self.rzx_lines.set_x_y_data(X, Y)

            X = Y
            Y = [-p.z.imag for p in points]
            self.nyq_lines.set_x_y_data(X, Y)

            updated = True

            if self.auto_snap:
                self.rzx_plot.snap_bounds()
                self.phi_plot.snap_bounds()
                self.nyq_plot.snap_bounds()
                self.auto_snap = False

        return updated

    def sweep_started_callback(self, _tc, pt, _duration_ms, hires, f0, f1):
        with self.data_lock:
            self.end_sweep_time = pt.t_timeout
            freq_str            = '[%.2f - %.2f]' % (f0, f1)
            if hires:
                self.sweep_prefix = 'Sweeping Hires %s' % freq_str
            else:
                self.sweep_prefix = 'Searching %s' % freq_str
            self.mark_dirty()

    def sweep_callback(self, _tc, pt, t0_ns, duration_ms, points, fw_fit,
                       hires, _temp_freq, _temp_c):
        with self.data_lock:
            if pt.sweep_iter > 1:
                fw_fit._gl_time_ns  = t0_ns + duration_ms * 1000
                fw_fit._gl_dtime_ns = fw_fit._gl_time_ns - pt.start_time_ns
                self.fits.append(fw_fit)
            self.points = points
            self.mark_dirty()

        if pt.sweep_iter > 1 and self.csv:
            self.csv.write('%u,%u,%.3f,%.3f\n' %
                           (fw_fit._gl_time_ns, fw_fit._gl_dtime_ns,
                            fw_fit.temp_c, fw_fit.peak_hz))
            self.csv.flush()


def main(args):
    if args.output_csv:
        csv = open(  # pylint: disable=R1732
                args.output_csv, 'a', encoding='utf8')
        csv.write('time_ns,dt,temp_c,peak_hz\n')
        csv.flush()
    else:
        csv = None

    dev    = xtalx.z_sensor.find_one(serial_number=args.sensor)
    tc     = xtalx.z_sensor.make(dev, yield_Y=True)
    ww     = WATWindow(tc, csv)
    a      = tc.parse_amplitude(None)
    pt     = xtalx.z_sensor.PeakTracker(tc, a, PT_F0, PT_F1, PT_DF, PT_NFREQS,
                                        PT_SEARCH_TIME_SEC, PT_SWEEP_TIME_SEC,
                                        settle_ms=PT_SETTLE_TIME_MS,
                                        delegate=ww, enable_chirp=ENABLE_CHIRP)
    pt.start_threaded()

    try:
        glotlib.interact()
    except KeyboardInterrupt:
        print()
    finally:
        pt.stop_threaded()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--output-csv', '-o')
    parser.add_argument('--sensor', '-s')
    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# Copyright (c) 2020-2024 by Phase Advanced Sensor Systems Corp.
import argparse

import xtalx.p_sensor


def main(args):
    d = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
    x = xtalx.p_sensor.make(d)
    while True:
        m = x.read_measurement()
        print(m.tostring(args.verbose))


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    parser.add_argument('--verbose', '-v', action='store_true')
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
from .tcsc import TCSC


class TCSC_U5(TCSC):
    DAC_MAX  = 4096
    ADC_MAX  = 16384
    ADC_KEYS = ('PROBEA', 'SIGIN')



# Copyright (c) 2022-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import math

import scipy.optimize
import numpy as np


def lorentz_function(x, A, x0, W):
    '''
    Lorentz function with no background term and no Y offset.
    '''
    return (A / math.pi) * (W / ((x - x0)**2 + W**2))


class Lorentzian:
    def __init__(self, A, x0, W, RR=0, X=None, Y=None):
        if A < 0 and W < 0:
            A = -A
            W = -W

        self.A  = A
        self.x0 = x0
        self.W  = W
        if X is None or Y is None:
            self.RR = RR
        else:
            self.RR = self._compute_RR(X, Y)

    @staticmethod
    def from_x_y(X, Y, A, x0, w):
        cf   = scipy.optimize.curve_fit(lorentz_function, X, Y, p0=(A, x0, w))
        popt = cf[0]
        return Lorentzian(popt[0], popt[1], popt[2], X=X, Y=Y)

    def __call__(self, x):
        return lorentz_function(x, self.A, self.x0, self.W)

    def __repr__(self):
        return ('Lorentzian(A=%f, x0=%f, W=%f, RR=%f)' %
                (self.A, self.x0, self.W, self.RR))

    def _compute_RR(self, X, Y):
        Yf     = np.array([self(x) for x in X])
        ss_res = np.sum((Y - Yf)**2)
        ss_tot = np.sum((Y - np.mean(Y))**2)
        return 1 - (ss_res / ss_tot)



import time


class DateFormat:
    '''
    Takes a time string in *local* time and converts it to a POSIX
    timestamp.
    '''
    def __init__(self, date_fmt='%m/%d/%Y %H:%M:%S'):
        self.date_fmt = date_fmt

    def __call__(self, v):
        return time.mktime(time.strptime(v, self.date_fmt))


class Decoder:
    def __init__(self, f, fmt_table):
        self.fmt_table = fmt_table

        hdr_line = f.readline().strip()
        assert '"' not in hdr_line
        assert "'" not in hdr_line
        self.headers = hdr_line.split(',')
        req_hdrs = set(fmt_table.keys())
        got_hdrs = set(self.headers)
        missing_hdrs = req_hdrs - got_hdrs
        if missing_hdrs:
            raise Exception('Missing columns: %s' % missing_hdrs)

        self.points = []
        for l in f.readlines():
            l = l.strip()
            if not l:
                continue
            if l == hdr_line:
                continue
            if l.startswith('#'):
                continue

            assert '"' not in l
            assert "'" not in l
            fields = l.split(',')
            point  = {}
            for header, field in zip(self.headers, fields):
                fmt = self.fmt_table.get(header)
                if not fmt:
                    continue
                if field == '':
                    point[header] = None
                else:
                    point[header] = fmt(field)
            self.points.append(point)

    def get_column(self, name):
        return [p[name] for p in self.points]



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import threading
import time

import xtalx.z_sensor


TEMP_RUNNING = True


def temp_thread(tc, interval_secs):
    tc.start_fixed_out(0, 0)
    tc.set_t_enable(True)

    t0_crystal_ticks, t0_cpu_ticks = tc.read_temp()
    while TEMP_RUNNING:
        time.sleep(interval_secs)
        t1_crystal_ticks, t1_cpu_ticks = tc.read_temp()

        dt = (t1_cpu_ticks - t0_cpu_ticks) / tc.CPU_FREQ
        dcrystal = (t1_crystal_ticks - t0_crystal_ticks) & 0xFFFFFFFF
        t0_crystal_ticks, t0_cpu_ticks = t1_crystal_ticks, t1_cpu_ticks

        if dt == 0:
            tc.warn('Temp crystal does not appear to be ticking.')
            continue

        temp_hz = dcrystal * 8 / dt

        temp_c, _, _ = tc.eval_freqs(temp_hz, 0, 0)
        tc.info('temp_hz %s T %s' % (temp_hz, temp_c))


def main(args):
    global TEMP_RUNNING

    dev = xtalx.z_sensor.find_one(serial_number=args.sensor)
    tc  = xtalx.z_sensor.make(dev, verbose=args.verbose)
    t   = threading.Thread(target=temp_thread, args=(tc, args.interval_secs))
    t.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print()
    finally:
        TEMP_RUNNING = False
        t.join()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--sensor', '-s')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--interval-secs', '-i', type=float, default=0.1)
    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    _main()



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import random
import cmath
import math
import time
import struct
from enum import IntEnum

import usb.core
import btype
import numpy as np

from .tincan import TinCan
from .scope_data import ScopeData


class SineFitPhasor(btype.Struct):
    x              = btype.float64_t()
    y              = btype.float64_t()
    offset         = btype.float64_t()
    RR             = btype.float64_t()
    _EXPECTED_SIZE = 32


class SampleHeader(btype.Struct):
    sfp            = btype.Array(SineFitPhasor(), 2)
    dds_skip       = btype.uint32_t()
    adc_skip       = btype.uint32_t()
    nsamples       = btype.uint32_t()
    t_ms           = btype.uint32_t()
    tag            = btype.uint16_t()
    rsrv           = btype.uint16_t()
    isr_cycles     = btype.Array(btype.uint32_t(), 3)
    _EXPECTED_SIZE = 96


class AutoChirpHeader(btype.Struct):
    nchirps         = btype.uint32_t()
    bin0            = btype.uint32_t()
    bin1            = btype.uint32_t()
    rsrv            = btype.uint32_t()
    bin_width       = btype.float64_t()
    A               = btype.float64_t()
    x0              = btype.float64_t()
    W               = btype.float64_t()
    _EXPECTED_SIZE  = 48


class AutoChirpResult:
    def __init__(self, hdr, bins):
        self.nchirps   = hdr.nchirps
        self.bin0      = hdr.bin0
        self.bin1      = hdr.bin1
        self.nbins     = hdr.bin1 - hdr.bin0 + 1
        self.bin_width = hdr.bin_width
        self.f0        = hdr.bin0 * hdr.bin_width
        self.f1        = hdr.bin1 * hdr.bin_width
        self.A         = hdr.A
        self.x0        = hdr.x0
        self.W         = hdr.W
        self.hdr       = hdr
        self.bins      = bins / hdr.nchirps


class StartCalPayload(btype.Struct):
    sigout          = btype.uint32_t()
    bias            = btype.uint32_t()
    _EXPECTED_SIZE  = 8


class EvalFreqsPayload(btype.Struct):
    temp_hz         = btype.float64_t()
    center_hz       = btype.float64_t()
    width_hz        = btype.float64_t()
    _EXPECTED_SIZE  = 24


class EvalFreqsResponse(btype.Struct):
    opcode                  = btype.uint16_t()
    tag                     = btype.uint16_t()
    status                  = btype.uint32_t()
    flags                   = btype.uint32_t()
    rsrv                    = btype.uint32_t()
    temp_c                  = btype.float64_t()
    density_g_per_ml        = btype.float64_t()
    viscosity_cp            = btype.float64_t()
    _EXPECTED_SIZE          = 40


class Status(IntEnum):
    OK              = 0
    BAD_OPCODE      = 1
    BAD_CMD_LENGTH  = 2
    ABORTED         = 3
    BUSY            = 4
    FAILED          = 5

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            s = '%s' % Status(rsp.status)
        except ValueError:
            s = '%u' % rsp.status
        return s


class CommandException(Exception):
    def __init__(self, rsp):
        super().__init__(
            'Command exception: %s (%s)' % (Status.rsp_to_status_str(rsp), rsp))
        self. rsp = rsp


class Opcode(IntEnum):
    GET_INFO        = 1
    START_SCOPER    = 2
    START_SWEEPER   = 4
    START_FIXED_OUT = 6
    SET_T_ENABLE    = 8
    READ_TEMP       = 9
    FIT_POINTS      = 10
    EVAL_FREQS      = 14
    GEN_HIRES_FREQS = 15
    GET_EINFO       = 16
    AUTO_CHIRP      = 17
    BAD_OPCODE      = 0xCCCC


class DriveType(IntEnum):
    UNKNOWN_DRIVE  = 0
    EXTERNAL_DRIVE = 1
    INTERNAL_DRIVE = 2


DRIVE_TYPE_MAP = {
    DriveType.UNKNOWN_DRIVE  : 'Unknwon',
    DriveType.EXTERNAL_DRIVE : 'External',
    DriveType.INTERNAL_DRIVE : 'Internal',
}


class ResetReason(IntEnum):
    UNKNOWN         = 0
    POWER_ON_RESET  = 1
    SOFTWARE_RESET  = 2
    STANDBY_RESET   = 3
    NRST_PIN        = 4


class CommandHeader(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    rsrv            = btype.uint32_t()
    _EXPECTED_SIZE  = 8


class StartScoperPayload(btype.Struct):
    dds_skip        = btype.uint32_t()
    amplitude       = btype.uint32_t()
    _EXPECTED_SIZE  = 8


class AutoChirpPayload(btype.Struct):
    skip0           = btype.uint32_t()
    skip1           = btype.uint32_t()
    amplitude       = btype.uint32_t()
    _EXPECTED_SIZE  = 12


class SetTEnablePayload(btype.Struct):
    enabled         = btype.uint32_t()
    _EXPECTED_SIZE  = 4


class FitCommandPayload(btype.Struct):
    flags           = btype.uint32_t()
    rsrv            = btype.Array(btype.uint8_t(), 12)
    temp_hz         = btype.float64_t()
    _EXPECTED_SIZE  = 24


class GenHiresFreqsPayload(btype.Struct):
    f0              = btype.float64_t()
    width           = btype.float64_t()
    N               = btype.uint32_t()
    _EXPECTED_SIZE  = 20


class Response(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    params          = btype.Array(btype.uint32_t(), 12)
    _EXPECTED_SIZE  = 56


class GetInfoResponse:
    def __init__(self, rsp, tc):
        self.hclk                 = rsp.params[0]
        self.dclk_divisor         = rsp.params[1] >> 16
        self.aclk_divisor         = rsp.params[1] & 0xFFFF
        self.cmd_buf_len          = rsp.params[2] >> 10
        self.f_hs_mhz             = rsp.params[2] & 0x03FF
        self.dclk                 = self.hclk / self.dclk_divisor
        self.aclk                 = self.hclk / self.aclk_divisor
        self.drive_type           = DriveType(rsp.params[4] & 0xFF)
        self.reset_reason         = ResetReason((rsp.params[4] >> 8) & 0xFF)
        self.cal_params           = rsp.params[4] >> 16
        self.reset_csr            = rsp.params[5]
        self.reset_sr1            = rsp.params[6]
        self.max_sweep_entries    = rsp.params[7] & 0xFFFF
        self.cal_dac_amplitude    = rsp.params[7] >> 16
        self.electronics_cal_date = rsp.params[8]
        self.crystal_cal_date     = rsp.params[9]
        self.air_f0               = rsp.params[10] / 1000
        self.air_fwhm             = rsp.params[11] / 1000

        if tc.fw_version < 0x106:
            self.nresets       = rsp.params[3]
            self.dv_nominal_hz = 32768
        else:
            self.nresets       = (rsp.params[3] >> 24) & 0xFF
            self.dv_nominal_hz = rsp.params[3] & 0xFFFFFF

    def have_temp_cal(self):
        return self.cal_params & (1 << 0)

    def get_drive_type(self):
        return DRIVE_TYPE_MAP.get(self.drive_type, 'Really Unknown')


class GetEInfoResponse(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    r_source        = btype.float64_t()
    r_feedback      = btype.float64_t()
    dac_to_v_coefs  = btype.Array(btype.float64_t(), 2)
    adc_to_v_coefs  = btype.Array(btype.float64_t(), 2)
    _EXPECTED_SIZE  = 56


class ReadTempResponse(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    status          = btype.uint32_t()
    cpu_ticks       = btype.uint64_t()
    osc_ticks       = btype.uint32_t()
    _EXPECTED_SIZE  = 20


class FitPointsResponse(btype.Struct):
    opcode                  = btype.uint16_t()
    tag                     = btype.uint16_t()
    status                  = btype.uint32_t()
    fit_flags               = btype.uint16_t()
    fit_niter               = btype.uint8_t()
    fit_status              = btype.int8_t()
    temp_c                  = btype.float32_t()
    fit_RR                  = btype.float32_t()
    strength                = btype.float32_t()
    density_g_per_ml        = btype.float64_t()
    viscosity_cp            = btype.float64_t()
    peak_hz                 = btype.float64_t()
    peak_fwhm               = btype.float64_t()
    _EXPECTED_SIZE          = 56


class SweepFit:
    def __init__(self, fpr, dt, temp_hz):
        self.status    = fpr.fit_status
        self.flags     = fpr.fit_flags
        self.niter     = fpr.fit_niter
        self.RR        = fpr.fit_RR
        self.peak_hz   = fpr.peak_hz
        self.peak_fwhm = fpr.peak_fwhm
        self.strength  = fpr.strength
        self.dt        = dt
        self.temp_hz   = temp_hz

        if self.flags & (1 << 0):
            self.temp_c = fpr.temp_c
        else:
            self.temp_c = None

        if self.flags & (1 << 1):
            self.density_g_per_ml = fpr.density_g_per_ml
        else:
            self.density_g_per_ml = None

        if self.flags & (1 << 2):
            self.viscosity_cp = fpr.viscosity_cp
        else:
            self.viscosity_cp = None


class StartSweepPayload(btype.Struct):
    nfreqs              = btype.uint32_t()
    amplitude           = btype.uint32_t()
    double_precision    = btype.uint32_t()
    _EXPECTED_SIZE      = 12


class SweepEntry(btype.Struct):
    dds_skip        = btype.uint32_t()
    ndiscards       = btype.uint16_t()
    nbufs           = btype.uint16_t()
    _EXPECTED_SIZE  = 8


class SweepResult(btype.Struct):
    real            = btype.Array(btype.float64_t(), 2)
    imag            = btype.Array(btype.float64_t(), 2)
    offset          = btype.Array(btype.float64_t(), 2)
    RR              = btype.Array(btype.float64_t(), 2)
    _EXPECTED_SIZE  = 64


class ParsedSweepResult:
    def __init__(self, sweep_result, rf, freq, nbufs, yield_Y):
        self._sweep_result = sweep_result

        phasors = [sweep_result.real[0] + sweep_result.imag[0] * 1j,
                   sweep_result.real[1] + sweep_result.imag[1] * 1j]

        self.amplitude = [abs(p) for p in phasors]
        self.phase     = [cmath.phase(p) for p in phasors]
        self.RR        = list(sweep_result.RR)
        self.nbufs     = nbufs
        self.f         = freq

        probea = phasors[0]
        sigin  = phasors[1]
        self.Z = -probea * rf / sigin
        self.Y = 1 / self.Z
        self.z = self.Y if yield_Y else self.Z


class SweepData:
    def __init__(self, sweep_results, sweep_params, rf, yield_Y):
        self.results = []
        for sr, (freq, nbufs) in zip(sweep_results, sweep_params):
            self.results.append(ParsedSweepResult(sr, rf, freq, nbufs, yield_Y))


class TCSC(TinCan):
    CMD_EP   = 0x01
    RSP_EP   = 0x82
    SCOPE_EP = 0x83

    DAC_MAX  = None
    ADC_MAX  = None
    ADC_KEYS = ()

    def __init__(self, usb_dev, **kwargs):
        super().__init__(**kwargs)

        self.usb_dev = usb_dev

        try:
            self.serial_num = usb_dev.serial_number
            self.git_sha1   = usb.util.get_string(usb_dev, 4)
            self.fw_version = usb_dev.bcdDevice
        except ValueError as e:
            if str(e) == 'The device has no langid':
                raise Exception(
                    'Device has no langid, ensure running as root!') from e
        self.info('Controlling sensor %s with firmware 0x%X (%s).' %
                  (self.serial_num, self.fw_version, self.git_sha1))

        if self.fw_version < 0x101:
            raise Exception('Firmware version of 0x%X not supported.' %
                            self.fw_version)

        self.tag = random.randint(1, 0xFFFF)

        self._set_configuration(usb_dev, 0x15)
        self._synchronize()

        self.ginfo    = self._get_info()
        self.einfo    = self._get_einfo()
        self.CPU_FREQ = self.ginfo.hclk

        self._sweep_params = None

    @staticmethod
    def _set_configuration(usb_dev, bConfigurationValue, force=False):
        cfg = None
        try:
            cfg = usb_dev.get_active_configuration()
        except usb.core.USBError as e:
            if e.strerror != 'Configuration not set':
                raise

        if (cfg is None or cfg.bConfigurationValue != bConfigurationValue or
                force):
            usb.util.dispose_resources(usb_dev)
            usb_dev.set_configuration(bConfigurationValue)

    def _send_abort(self):
        self.usb_dev.write(self.CMD_EP, b'')

    def _alloc_tag(self):
        tag      = self.tag
        self.tag = 1 if self.tag == 0xFFFF else self.tag + 1
        return tag

    def _synchronize(self):
        self._send_abort()
        tag  = self._alloc_tag()
        hdr  = CommandHeader(opcode=Opcode.BAD_OPCODE, tag=tag)
        data = hdr.pack() + bytes(48)
        junk = 0
        while True:
            self.usb_dev.write(self.CMD_EP, data)
            rsp = self.usb_dev.read(self.RSP_EP, 64, timeout=100)
            if len(rsp) == 56:
                break

            junk += len(rsp)

        self.usb_dev.write(self.CMD_EP, data)
        data = self.usb_dev.read(self.RSP_EP, 64)
        assert len(data) == 56

        rsp = Response.unpack(data)
        assert rsp.tag    == tag
        assert rsp.opcode == Opcode.BAD_OPCODE
        assert rsp.status == Status.BAD_OPCODE

        junk_len = 0
        try:
            while True:
                junk_len += len(self.usb_dev.read(self.SCOPE_EP, 64,
                                                  timeout=100))
        except usb.core.USBTimeoutError:
            pass

        return junk

    def _send_command(self, opcode, params, bulk_data, timeout):
        tag  = self._alloc_tag()
        hdr  = CommandHeader(opcode=opcode, tag=tag)
        data = hdr.pack() + params + bytes(48 - len(params)) + bulk_data
        size = self.usb_dev.write(self.CMD_EP, data, timeout=timeout)
        assert size == len(data)
        if (len(data) % 64) == 0:
            if len(data) < self.ginfo.cmd_buf_len:
                self.usb_dev.write(self.CMD_EP, b'')
        return tag

    def _recv_response(self, tag, timeout, cls=Response):
        data = self.usb_dev.read(self.RSP_EP, Response._STRUCT.size,
                                 timeout=timeout)
        assert len(data) == Response._STRUCT.size
        rsp = cls.unpack_from(data)
        assert rsp.tag == tag

        if rsp.status != Status.OK:
            rsp.opcode = Opcode(rsp.opcode)
            raise CommandException(rsp)

        return rsp

    def _exec_command(self, opcode, params=b'', bulk_data=b'', timeout=1000,
                      cls=Response):
        tag = self._send_command(opcode, params, bulk_data, timeout)
        return self._recv_response(tag, timeout, cls=cls)

    def _get_info(self):
        return GetInfoResponse(self._exec_command(Opcode.GET_INFO), self)

    def _get_einfo(self):
        return self._exec_command(Opcode.GET_EINFO, cls=GetEInfoResponse)

    def fft_limit(self):
        return self.ginfo.aclk / 4

    def ms_samples(self):
        return int(self.ginfo.aclk / 1000)

    def cal_dac_amp(self):
        return self.ginfo.cal_dac_amplitude

    def v_to_adc(self, v):
        '''
        Return the ADC value that would be measured for the given voltage.

        Note that ADC values are signed because the TCSC FW shifts the values
        down by ADC_MAX / 2 to center on the origin for stability in the
        floating-point math.
        '''
        if not self.ginfo.electronics_cal_date:
            return None
        return ((v - self.einfo.adc_to_v_coefs[0]) /
                self.einfo.adc_to_v_coefs[1])

    def adc_to_v(self, adc):
        '''
        Return the voltage that would generate the given ADC value.

        Note that ADC values are signed because the TCSC FW shifts the values
        down by ADC_MAX / 2 to center on the origin for stability in the
        floating-point math.
        '''
        if not self.ginfo.electronics_cal_date:
            return None
        return (self.einfo.adc_to_v_coefs[0] +
                adc * self.einfo.adc_to_v_coefs[1])

    def v_to_dac(self, v):
        '''
        Returns the DAC value corresponding with the specified output voltage
        if the bias voltage is set to half VREF (i.e. 2048).  Otherwise, this
        value is only suitable for relative voltage calculations.
        '''
        if not self.ginfo.electronics_cal_date:
            return None
        return ((v - self.einfo.dac_to_v_coefs[0]) /
                self.einfo.dac_to_v_coefs[1])

    def a_to_dac(self, a):
        '''
        Given an amplitude in volts, return the equivalent amplitude in DAC
        units.  Note that an amplitude is a relative voltage measurement and
        only takes into account the gain; the offset will be around the bias
        voltage but is not calibrated.
        '''
        if not self.ginfo.electronics_cal_date:
            return None
        return -a / self.einfo.dac_to_v_coefs[1]

    def dac_to_a(self, dac):
        '''
        Given an amplitude in DAC units, return the equivalent amplitude in
        volts.  Note that an amplitude is a relative voltage measurement and
        only takes into account the gain; the offset will be around the bias
        voltage but is not calibrated.
        '''
        if not self.ginfo.electronics_cal_date:
            return None
        return -dac * self.einfo.dac_to_v_coefs[1]

    def parse_amplitude(self, amplitude):
        if amplitude is None:
            amplitude = self.cal_dac_amp()
            if amplitude is None:
                raise Exception("Calibration page doesn't include the DAC "
                                "voltage under which the calibration was "
                                "performed, must specify --amplitude manually.")
            return amplitude

        if amplitude.upper().endswith('V'):
            volts = float(amplitude[:-1])
            amplitude = self.a_to_dac(volts)
            if amplitude is None:
                raise Exception("Calibration page doesn't have required "
                                "voltage-to-DAC information to use amplitudes "
                                "in Volts.")
            return round(amplitude)

        return int(amplitude)

    def _read_samples(self):
        data    = self.usb_dev.read(self.SCOPE_EP, 768 * 1024, timeout=3000)
        hdr     = SampleHeader.unpack(data[:SampleHeader._STRUCT.size])
        assert len(data) == SampleHeader._STRUCT.size + hdr.nsamples * 2
        samples = np.frombuffer(data, dtype='<i2',
                                offset=SampleHeader._STRUCT.size)
        assert len(samples) == hdr.nsamples

        w    = 2 * math.pi * hdr.adc_skip / 2**32
        freq = self.ginfo.dclk * hdr.dds_skip / 2**32
        sd   = ScopeData(freq, w, hdr=hdr, data=data)
        for i, name in enumerate(self.ADC_KEYS):
            offset = hdr.sfp[i].offset + self.ADC_MAX / 2
            X      = np.arange(i, len(samples), 2)
            Y      = samples[i::2] + self.ADC_MAX / 2
            clk    = self.ginfo.aclk / 2
            sd.add_signal(X, Y, clk, name, offset=offset)

        return sd, hdr

    def send_scope_cmd(self, frequency, amplitude):
        dds_skip = int(round(2**32 * (frequency / self.ginfo.dclk)))
        return self._exec_command(
            Opcode.START_SCOPER,
            StartScoperPayload(dds_skip=dds_skip, amplitude=amplitude).pack())

    def sample_scope_sync(self, **_kwargs):
        sd, hdr = self._read_samples()

        for si, sfp in zip(sd.sig_info, hdr.sfp):
            si.phasor = sfp.x + sfp.y * 1j
            si.RR     = sfp.RR

        return sd

    def send_auto_chirp_cmd(self, f0, f1, amplitude):
        skip0 = int(round(2**32 * (f0 / self.ginfo.dclk)))
        skip1 = int(round(2**32 * (f1 / self.ginfo.dclk)))
        return self._exec_command(
            Opcode.AUTO_CHIRP,
            AutoChirpPayload(skip0=skip0, skip1=skip1,
                             amplitude=amplitude).pack())

    def sample_auto_chirp_sync(self):
        data = self.usb_dev.read(self.SCOPE_EP, 128*1024, timeout=3000)
        assert len(data) == 16432

        hdr  = AutoChirpHeader.unpack(data[:AutoChirpHeader._STRUCT.size])
        bins = np.frombuffer(data, dtype='<f',
                             offset=AutoChirpHeader._STRUCT.size)
        return AutoChirpResult(hdr, bins)

    def sweep_async(self, amplitude, freq_tuples, ndiscards=2):
        assert 2 <= len(freq_tuples) <= self.ginfo.max_sweep_entries

        self._sweep_params = []
        params = StartSweepPayload(nfreqs=len(freq_tuples),
                                   amplitude=amplitude,
                                   double_precision=True).pack()
        bulk_data = b''
        for i, (f, nbufs) in enumerate(freq_tuples):
            dds_skip = int(round(2**32 * (f / self.ginfo.dclk)))
            freq     = self.ginfo.dclk * dds_skip / 2**32
            bulk_data += SweepEntry(dds_skip=dds_skip,
                                    ndiscards=ndiscards if i == 0 else 5,
                                    nbufs=nbufs).pack()
            self._sweep_params.append((freq, nbufs))

        rsp = self._exec_command(Opcode.START_SWEEPER, params, bulk_data)
        sleep_ms = rsp.params[0]
        wakeup_time = time.localtime(time.time() + sleep_ms / 1000)

        self.info('Sweeping %u frequencies from %.3f-%.3f Hz.  Sensor '
                  'recommends sleep of %u ms until %s.' %
                  (len(freq_tuples), freq_tuples[0][0], freq_tuples[-1][0],
                   sleep_ms, time.asctime(wakeup_time)))

        return sleep_ms / 1000, sleep_ms

    def read_sweep_data(self):
        size = SweepResult._STRUCT.size * len(self._sweep_params)
        data = self.usb_dev.read(self.RSP_EP, size)
        assert len(data) == size

        results = []
        while data:
            sr = SweepResult.unpack(data[:SweepResult._STRUCT.size])
            sr.offset[0] += self.ADC_MAX / 2
            sr.offset[1] += self.ADC_MAX / 2
            results.append(sr)
            data = data[SweepResult._STRUCT.size:]

        return SweepData(results, self._sweep_params, self.einfo.r_feedback,
                         self.yield_Y)

    def get_sweep_fit(self, temp_hz):
        flags = 0
        if self.yield_Y:
            flags |= (1 << 0)

        t0 = time.time_ns()
        rsp = self._exec_command(Opcode.FIT_POINTS,
                                 FitCommandPayload(flags=flags,
                                                   temp_hz=temp_hz).pack(),
                                 timeout=20000, cls=FitPointsResponse)
        t1 = time.time_ns()

        return SweepFit(rsp, t1 - t0, temp_hz)

    def start_fixed_out(self, sigout, bias):
        '''
        Generate a fixed voltage output, setting the output op-amp's positive
        input to bias and negative input to sigout.  This is typically used
        for calibration purposes, but setting sigout=0 and bias=0 can also be
        used to quiesece the sensor by grounding the output.
        '''
        return self._exec_command(
            Opcode.START_FIXED_OUT,
            StartCalPayload(sigout=sigout, bias=bias).pack())

    def set_t_enable(self, enabled):
        '''
        Enable or disable the temperature oscillators to reduce noise on the
        density/viscosity measurement.
        '''
        params = SetTEnablePayload(enabled=enabled).pack()
        self._exec_command(Opcode.SET_T_ENABLE, params)

    def read_temp(self):
        rsp = self._exec_command(Opcode.READ_TEMP, b'', cls=ReadTempResponse)
        return rsp.osc_ticks, rsp.cpu_ticks

    def eval_freqs(self, temp_hz, center_hz, width_hz):
        '''
        Get the sensor to manually convert a set of frequencies into
        temperature, density and viscosity values.  To only convert
        temperature, set center_hz=0 and width_hz=0.
        '''
        params = EvalFreqsPayload(temp_hz=temp_hz, center_hz=center_hz,
                                  width_hz=width_hz).pack()
        rsp    = self._exec_command(Opcode.EVAL_FREQS, params,
                                    cls=EvalFreqsResponse)
        temp_c = rsp.temp_c if rsp.flags & 1 else None
        if center_hz and width_hz:
            density_g_per_ml = rsp.density_g_per_ml if rsp.flags & 2 else None
            viscosity_cp     = rsp.viscosity_cp if rsp.flags & 4 else None
        else:
            density_g_per_ml = None
            viscosity_cp     = None
        return (temp_c, density_g_per_ml, viscosity_cp)

    def gen_hires_freqs(self, f0, width, N):
        '''
        Generate a list of hi-resolution frequencies to be measured, centered
        on the peak at f0 of the specified width, with N specifying the number
        of frequencies in each wing.  A total of 2N + 1 frequencies will be
        returned.
        '''
        params = GenHiresFreqsPayload(f0=f0, width=width, N=N).pack()
        self._send_command(Opcode.GEN_HIRES_FREQS, params, b'', 1000)

        N    = 2*N + 1
        size = 8*N
        data = self.usb_dev.read(self.RSP_EP, size, timeout=1000)
        return struct.unpack('<%ud' % N, data)



#!/usr/bin/env python3
# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import usb.util

import xtalx.z_sensor


def main():
    for s in xtalx.z_sensor.find():
        print('******************')
        print('  Product: %s' % s.product)
        print('Sensor SN: %s' % s.serial_number)
        print(' git SHA1: %s' % usb.util.get_string(s, 4))
        print('  Version: 0x%04X' % s.bcdDevice)


if __name__ == '__main__':
    main()



# Copyright (c) 2020-2023 Phase Advanced Sensor Systems, Inc.
# All rights reserved.


class InvalConfigException(Exception):
    pass


class Config:
    def __init__(self, lines, required_keys):
        self.kv_pairs = {}
        rks = set(required_keys)
        for l in lines:
            if l.startswith('#'):
                continue
            k, v = l.split()
            if k in self.kv_pairs:
                raise InvalConfigException('Key %s repeated.' % k)
            self.kv_pairs[k] = v
            rks.discard(k)
        if rks:
            raise InvalConfigException('Missing keys: %s' % rks)

    def has_keys(self, required_keys):
        for k in required_keys:
            if k not in self.kv_pairs:
                return False
        return True

    def __getattr__(self, k):
        return self.kv_pairs[k]



#!/usr/bin/env python3
# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import threading
import cmath
import math
import time

import numpy
import scipy.fft

import glotlib

import xtalx.z_sensor


class ScopeWindow(glotlib.Window):
    def __init__(self, tc, a, f, hide_fits, verbose):
        super().__init__(900, 700, msaa=4)

        self.data_lock   = threading.Lock()
        self.data        = None
        self.running     = False
        self.tc         = tc
        self.frequency   = f
        self.amplitude   = a
        self.hide_fits   = hide_fits
        self.verbose     = verbose
        self.num_samples = 0

        self.fig         = None
        self.title       = None
        self.adc         = None
        self.adc_fit     = None
        self.fft         = None

        ax = self.add_plot((8, 1, (1, 5)),
                           limits=(0, 0, self.tc.ms_samples(),
                                   self.tc.ADC_MAX))
        self.adc = [ax.add_steps() for k in self.tc.ADC_KEYS]
        if not self.hide_fits:
            self.adc_fit = [ax.add_lines() for _ in self.tc.ADC_KEYS]

        self.fft = []

        ax = self.add_plot(817, limits=(-1000, -1000,
                                        1000 + self.tc.fft_limit(),
                                        1000 + self.tc.ADC_MAX / 2))
        self.fft.append(ax.add_steps())

        ax = self.add_plot(818, sharex=ax,
                           limits=(-1000, -1000,
                                   1000 + self.tc.fft_limit(),
                                   1000 + self.tc.ADC_MAX / 2))
        self.fft.append(ax.add_steps())

    def start(self):
        self.running = True
        threading.Thread(target=self.usb_thread).start()

    def stop(self):
        self.running = False

    def update_geometry(self, _t):
        with self.data_lock:
            sd        = self.data
            self.data = None
        if sd is None:
            return False

        w = sd.w
        X = [i/10 for i in range(self.tc.ms_samples()*10)]
        for i, si in enumerate(sd.sig_info):
            self.adc[i].set_x_y_data(si.X, si.Y)

            N  = len(si.Y)
            yf = scipy.fft.fft(si.Y)[1:N//2] / N * 2
            xf = scipy.fft.fftfreq(N, 1 / si.clk)[1:N//2]
            self.fft[i].set_x_y_data(xf, numpy.abs(yf))

            if not self.hide_fits:
                A, phase = cmath.polar(si.phasor)
                Y = [si.offset + A * math.sin(x*w + phase) for x in X]
                self.adc_fit[i].set_x_y_data(X, Y)

        self.num_samples += 1

        return True

    def usb_thread(self):
        total_len = len1 = 0

        self.tc.send_scope_cmd(self.frequency, self.amplitude)
        t0 = t1 = time.time()
        while self.running:
            time.sleep(0.02)
            sd = self.tc.sample_scope_sync()

            new_len = total_len + len(sd.data)
            if total_len // (1024*1024) != new_len // (1024*1024):
                t    = time.time()
                dt0  = t - t0
                dt1  = t - t1
                dlen = new_len - len1
                t1   = t
                len1 = new_len
                self.tc.info(
                    '%u bytes in %s seconds (%f bytes/s, avg %f bytes/s) '
                    '[%u / %u / %u] [%u - %u] [%u - %u]' %
                    (new_len, dt0, dlen / dt1, new_len / dt0,
                     sd.hdr.isr_cycles[0], sd.hdr.isr_cycles[1],
                     sd.hdr.isr_cycles[2],
                     min(sd.sig_info[0].Y), max(sd.sig_info[0].Y),
                     min(sd.sig_info[1].Y), max(sd.sig_info[1].Y)))
            total_len = new_len

            with self.data_lock:
                self.data = sd
                self.mark_dirty()


def main(rv):
    dev = xtalx.z_sensor.find_one(serial_number=rv.sensor)
    tc  = xtalx.z_sensor.make(dev)

    tc.info('Using frequency of %s Hz.' % rv.freq)
    if rv.amplitude.upper().endswith('V'):
        volts = float(rv.amplitude[:-1])
        amplitude = round(tc.a_to_dac(volts))
        if amplitude is None:
            raise Exception("Calibration page doesn't have required voltage-"
                            "to-DAC information to use amplitudes in Volts.")

        tc.info('Using amplitude of %u = %fV' %
                (amplitude, tc.dac_to_a(amplitude)))
    else:
        amplitude = int(rv.amplitude)
        volts = tc.dac_to_a(amplitude)
        if volts is not None:
            tc.info('Using amplitude of %u = %fV' %
                    (amplitude, tc.dac_to_a(amplitude)))

    if not 0 <= amplitude <= 2000:
        raise Exception('Amplitude not in range 0 to 2000.')

    sw = ScopeWindow(tc, amplitude, rv.freq, rv.hide_fits, rv.verbose)
    sw.start()
    try:
        glotlib.interact()
    finally:
        sw.stop()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--freq', '-f', type=float, default=33000)
    parser.add_argument('--amplitude', '-a', default='0.1v')
    parser.add_argument('--verbose', '-v', action='store_true')
    parser.add_argument('--hide-fits', action='store_true')
    parser.add_argument('--sensor', '-s')
    rv = parser.parse_args()

    try:
        main(rv)
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2022-2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import time
from struct import unpack

from xtalx.tools.math import PolynomialFit1D, PolynomialFit2D

import xtalx.tools.serial


def hex_to_double(h):
    '''
    Converts a 64-bit value encoded as a hex string into a floating-point
    double value.
    '''
    h = bytes([int(h[i:i+2], 16) for i in range(14, -2, -2)])
    return unpack('<d', h)[0]


class XHTICommandException(Exception):
    def __init__(self, rsp):
        super().__init__('Unexpected response: "%s"' % rsp)
        self.rsp = rsp


class Measurement:
    def __init__(self, sensor, time_ns, pressure_psi, temp_c, pressure_freq,
                 temp_freq, lse_freq, clkcal_freq):
        self.sensor        = sensor
        self.time_ns       = time_ns
        self.pressure_psi  = pressure_psi
        self.temp_c        = temp_c
        self.pressure_freq = pressure_freq
        self.temp_freq     = temp_freq
        self.lse_freq      = lse_freq
        self.clkcal_freq   = clkcal_freq

    @staticmethod
    def _from_packet(sensor, time_ns, t_count, p_count, l_count, c_count):
        t_freq = sensor.refclk_hz * 26200 / t_count if t_count else 0
        p_freq = sensor.refclk_hz * 5000  / p_count if p_count else 0
        l_freq = sensor.refclk_hz * 3272  / l_count if l_count else 0
        c_freq = c_count * 32768 / 3272

        if sensor.poly_temp and t_freq:
            temp_c = sensor.poly_temp(t_freq)
        else:
            temp_c = None

        if sensor.poly_psi and p_freq and t_freq:
            psi = sensor.poly_psi(p_freq, t_freq)
        else:
            psi = None

        return Measurement(sensor, time_ns, psi, temp_c, p_freq, t_freq, l_freq,
                           c_freq)

    def tostring(self, verbose=False):
        s = '%s: ' % self.sensor
        if verbose:
            s += ('pf %s tf %s p %s t %s' %
                  (self.pressure_freq, self.temp_freq,
                   self.pressure_psi, self.temp_c))
        else:
            if self.pressure_psi is None:
                p = 'n/a'
            else:
                p = '%f' % self.pressure_psi

            if self.temp_c is None:
                t = 'n/a'
            else:
                t = '%f' % self.temp_c
            s += '%s PSI, %s C' % (p, t)

        return s

    def to_influx_point(self, time_ns=None, measurement='xtalx_data',
                        fields=None):
        time_ns = time_ns or time.time_ns()
        fields  = fields or {}
        p = {
            'measurement' : measurement,
            'time'        : time_ns,
            'tags'        : {'sensor' : self.sensor.serial_num},
            'fields'      : fields,
        }
        if self.pressure_psi is not None:
            fields['pressure_psi'] = float(self.pressure_psi)
        if self.temp_c is not None:
            fields['temp_c'] = float(self.temp_c)
        if self.pressure_freq is not None:
            fields['pressure_freq_hz'] = float(self.pressure_freq)
        if self.temp_freq is not None:
            fields['temp_freq_hz'] = float(self.temp_freq)
        return p


class XHTI:
    def __init__(self, intf, baudrate=None, spy=False):
        assert intf is not None

        if baudrate is None:
            baudrate = 57600

        self.intf   = intf
        self.serial = xtalx.tools.serial.from_intf(intf, spy=spy,
                                                   baudrate=baudrate,
                                                   timeout=1, exclusive=True)

        self.fw_name        = None
        self.platform       = None
        self.fw_version_str = None
        self.fw_version     = None
        self.git_sha1       = None
        self.gcc_version    = None
        self.serial_num     = None
        self.serial_date    = None
        self.refclk_hz      = None
        self.report_id      = None
        self.bias           = None
        self.p_startup_ms   = None
        self.poly_psi       = None
        self.poly_temp      = None
        self.startup_err    = None
        self._halt_yield    = True

        self._reset_and_synchronize()

    def __str__(self):
        return 'XHTI(%s)' % self.serial_num

    def _write(self, data):
        for c in data:
            self.serial.write(b'%c' % c)
            self.serial.flush()
            time.sleep(7 / 57600)

    def _readline(self):
        return self.serial.readline()

    def _flush_input(self):
        timeout = self.serial.timeout
        self.serial.timeout = 0
        while self.serial.read_all():
            pass
        self.serial.timeout = timeout

    def _exec_command(self, cmd):
        self._write(cmd + b'\r')

        l = self._readline()
        if not l.startswith(b'S: '):
            print('Failed: %s' % cmd)
            raise XHTICommandException(l)
        return l

    def _exec_long_command(self, cmd):
        self._write(cmd + b'\r')
        rsp = []
        while True:
            l = self._readline()
            if l == b'#\r\n':
                raise XHTICommandException(rsp)
            if l == b'=\r\n':
                return rsp

            rsp.append(l)

            if self.fw_version < 0x210 and l.startswith(b'E: '):
                raise XHTICommandException(rsp)

    def _reset_and_synchronize(self):
        self._flush_input()
        self._write(b'R\r')
        self._synchronize()

    def _handle_R_line(self, l):
        # Handle the Reset line.
        assert l.startswith(b'R: ')
        words               = l.split()
        self.fw_name        = words[1].decode()
        self.platform       = words[2].decode()
        self.fw_version_str = words[3].decode()

        parts           = self.fw_version_str.split('.')
        self.fw_version = ((int(parts[0]) << 8) |
                           (int(parts[1]) << 4) |
                           (int(parts[2]) << 0))

    def _handle_G_line(self, l):
        # Handle the Git SHA1 line.
        assert l.startswith(b'G: ')
        self.git_sha1 = l.split()[-1].decode()

    def _handle_c_line(self, l):
        # Handle the compiler version line.
        assert l.startswith(b'c: ')
        self.gcc_version = l[3:].decode()

    def _handle_I_line(self, l):
        # Handle the Identity line.
        assert l.startswith(b'I: ')
        words            = l.split()
        self.serial_num  = words[1].decode()
        self.serial_date = words[2].decode()

    def _handle_F_line(self, l):
        # Nothing to do.
        pass

    def _handle_P_line(self, l):
        # Handle the PLL line.
        assert l.startswith(b'P: Max PLL @ ')
        _, _, r = l.partition(b' @ ')
        words   = r.split()
        assert words[1] == b'Hz'
        self.refclk_hz = int(words[0])

    def _handle_e_line(self, l):
        # Handle the startup error line.
        assert l.startswith(b'e: 0x')
        _, err           = l.split()
        self.startup_err = int(err, 16)

    def _synchronize(self, timeout=10):
        # Nuke everything.
        self.fw_name        = None
        self.platform       = None
        self.fw_version_str = None
        self.fw_version     = None
        self.git_sha1       = None
        self.gcc_version    = None
        self.serial_num     = None
        self.serial_date    = None
        self.refclk_hz      = None
        self.report_id      = None
        self.bias           = None
        self.p_startup_ms   = None
        self.poly_psi       = None
        self.poly_temp      = None
        self.startup_err    = 0

        # Wait for the Reset line.
        t0 = time.time()
        while True:
            l = self._readline()
            if l.startswith(b'R: '):
                break
            if time.time() - t0 >= timeout:
                raise Exception('Reset failed.')
        self._handle_R_line(l)

        # Handle all other lines.
        while True:
            l = self._readline()
            if l in (b'=\r\n', b'#\r\n'):
                break
            if l.startswith(b'G: '):
                self._handle_G_line(l)
            if l.startswith(b'c: '):
                self._handle_c_line(l)
            elif l.startswith(b'I: '):
                self._handle_I_line(l)
            elif l.startswith(b'F: '):
                self._handle_F_line(l)
            elif l.startswith(b'P: '):
                self._handle_P_line(l)
            elif l.startswith(b'e: '):
                self._handle_e_line(l)

        # Get calibration info.
        hdr = self.read_calibration_header()
        _, self.report_id, self.bias, self.p_startup_ms = hdr

        # Finally, read polynomials.
        self.poly_psi  = self.read_pressure_polynomial()
        self.poly_temp = self.read_temperature_polynomial()

    def read_calibration_header(self):
        '''
        Returns the tuple (hse_freq_x10, report_id, bias, p_startup_ms).
        '''
        if self.fw_version < 0x216:
            rsp = self._exec_command(b'HDR')
        else:
            rsp = self._exec_long_command(b'HDR')
            assert len(rsp) == 1
            rsp = rsp[0]
        words = rsp[3:].split()

        d = dict(zip(words[::2], words[1::2]))
        for k in d:
            if k not in (b'RefClk', b'Id', b'Bias', b'PStartupMs', b'PLLClk'):
                raise Exception('Unrecognized response: %s' % rsp)

        ref_clk      = float(d[b'RefClk'])
        report_id    = int(d[b'Id'])
        p_startup_ms = int(d[b'PStartupMs'])
        bias         = int(d.get(b'Bias', 12053700))
        pll_clk      = int(d.get(b'PLLClk', self.refclk_hz))
        assert pll_clk == self.refclk_hz

        return ref_clk, report_id, bias, p_startup_ms

    def read_pressure_polynomial(self):
        try:
            lines = self._exec_long_command(b'PLP')
        except XHTICommandException:
            return None

        assert len(lines) >= 3
        order    = len(lines) - 3
        words    = lines[0].split(b',')
        x_domain = (hex_to_double(words[0]), hex_to_double(words[1]))
        words    = lines[1].split(b',')
        y_domain = (hex_to_double(words[0]), hex_to_double(words[1]))
        coefs    = []
        for i in range(order + 1):
            words = lines[2 + i].split(b',')
            assert len(words) == order + 1
            coefs = coefs + [hex_to_double(w) for w in words]
        return PolynomialFit2D.from_domain_coefs(x_domain, y_domain, coefs)

    def read_temperature_polynomial(self):
        try:
            lines = self._exec_long_command(b'PLT')
        except XHTICommandException:
            return None

        assert len(lines) == 2
        words    = lines[0].split(b',')
        x_domain = (hex_to_double(words[0]), hex_to_double(words[1]))
        words    = lines[1].split(b',')
        coefs    = [hex_to_double(w) for w in words]
        return PolynomialFit1D.from_domain_coefs(x_domain, coefs)

    def yield_measurements(self):
        self.serial.flushInput()
        self._write(b'CAL\r')
        self._halt_yield = False
        while not self._halt_yield:
            l = self._readline()
            if not l.startswith(b'M: '):
                continue

            time_ns = time.time_ns()

            words = l.split()
            if (words[1][0] != ord('T') or words[2][0] != ord('P') or
                    words[3][0] != ord('L') or words[4][0] != ord('C')):
                continue

            t_count = int(words[1][1:], 16)
            p_count = int(words[2][1:], 16)
            l_count = int(words[3][1:], 16)
            c_count = int(words[4][1:], 16)
            if t_count == 0xFFFFFFFF:
                t_count = 0
            if p_count == 0xFFFFFFFF:
                p_count = 0
            if l_count == 0xFFFFFFFF:
                l_count = 0
            if c_count == 0xFFFFFFFF:
                c_count = 0

            yield Measurement._from_packet(self, time_ns, t_count, p_count,
                                           l_count, c_count)



import math
import argparse

from xtalx.tools.math import XYSeries
from xtalx.tools.csv import Decoder


CSV_FORMAT = {
    'time'          : float,
    'temp_c'        : float,
    'pressure_psi'  : float,
}


def resample(path, period):
    with open(path, 'r', encoding='utf8') as f:
        d = Decoder(f, CSV_FORMAT)

    X = [p['time'] - d.points[0]['time'] for p in d.points]
    t_series = XYSeries(X, [p['temp_c'] for p in d.points])
    p_series = XYSeries(X, [p['pressure_psi'] for p in d.points])
    N = math.ceil(X[-1] / period)

    rt_series = XYSeries([], [])
    rp_series = XYSeries([], [])
    for i in range(N):
        t0 = i*period
        rt_series.append(t0, t_series.get_avg_value(t0, t0+period))
        rp_series.append(t0, p_series.get_avg_value(t0, t0+period))

    return rt_series, rp_series


def main(args):
    rt_series, rp_series = resample(args.input_file, args.period)
    assert len(rt_series.X) == len(rp_series.X)

    with open(args.output_file, 'x', encoding='utf8') as f:
        f.write('time,temp_c,pressure_psi\n')
        for i, _ in enumerate(rt_series.X):
            f.write('%.6f,%.2f,%.6f\n' % (rt_series.X[i],
                                          rt_series.Y[i],
                                          rp_series.Y[i]))


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-file', '-i', required=True)
    parser.add_argument('--output-file', '-o', required=True)
    parser.add_argument('--period', '-p', type=float, default=1.0)
    main(parser.parse_args())


if __name__ == '__main__':
    _main()



# Copyright (c) 2020-2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import math
import numpy


class PolynomialFit1D:
    def __init__(self, order, pf):
        self.order = order
        self.pf    = pf

    @staticmethod
    def from_domain_coefs(x_domain, coefs):
        order = len(coefs) - 1
        pf    = numpy.polynomial.polynomial.Polynomial(coefs, domain=x_domain)
        return PolynomialFit1D(order, pf)

    def __call__(self, x):
        return self.pf(x)


class PolynomialFit2D:
    def __init__(self, x_domain, y_domain, coefs):
        self.x_domain = x_domain
        self.y_domain = y_domain
        self.coefs    = coefs

    @staticmethod
    def from_domain_coefs(x_domain, y_domain, coefs):
        '''
        Given the x_domain, y_domain and a list of coefficients of the form:

            [x0y0, x1y0, x2y0, ..., xNy0,
             x0y1, x1y1, x2y1, ..., xNy1,
             ...
             x0yN, x1yN, x2yN, ..., xNyN]

        generate the fit polynomial for evaluation purposes.
        '''
        order = round(math.sqrt(len(coefs))) - 1
        return PolynomialFit2D(x_domain, y_domain,
                               numpy.reshape(coefs, (order + 1, order + 1)))

    @staticmethod
    def _mapped(p, domain):
        l = domain[0]
        h = domain[1]
        return 2*(p - l)/(h - l) - 1

    def __call__(self, x, y):
        x = self._mapped(x, self.x_domain)
        y = self._mapped(y, self.y_domain)
        return numpy.polynomial.polynomial.polyval2d(x, y, self.coefs)
